

/*************
*            *
*  SENTINEL  *
*            *
*************/


#ifndef WWWI_MIND_DEBUGPTR_H
#define WWWI_MIND_DEBUGPTR_H
#ifdef USE_DEBUG_POINTERS


/**********************
*                     *
*  COMPILER INCLUDES  *
*                     *
**********************/


#include <cassert>
#include <cstdlib>
#include <iostream>
#include <map>
#include <typeinfo>


/*********************
*                    *
*  PROJECT INCLUDES  *
*                    *
*********************/


#include "wwwi/mutex.h"
#include "wwwi/mutexholder.h"
#include "wwwi/singleton.h"


using std::abort;
using std::cerr;
using std::endl;
using std::map;
using WWWI::GetSingleton;
using WWWI::Mutex;
using WWWI::MutexHolder;


/*****************
*                *
*  CLASS REFMGR  *
*                *
*****************/


/*
 *  The RefMgr class is used as a singleton to coordinate references to a 
 *  pointer.  If the number of references drops to 0 before the pointer is
 *  deleted, abort() will be called to generate a core dump.  At program
 *  exit, a summary of allocations of objects of this type will be printed, 
 *  along with a list of unfreed objects (if any).  One RefMgr class is 
 *  constructed for each type of object being tracked.
 *
 *  Methods
 *
 *    RefMgr(void)
 *      Invoked by GetSingleton, the constructor initializes the new/delete
 *      counts for its specified object type.
 *
 *    Delete(const T *ci_tp)
 *      This method is called to indicate that the given pointer is being 
 *      deleted.  If the reference count is zero (or less), abort() is called,
 *      indicating that the object has been freed twice.  This method does
 *      not actually delete the object.
 *
 *    Down(const T *ci_tp)
 *      This method decrements the reference count.  If the reference count
 *      reaches zero, it means that no pointers remain that can free the
 *      object.  abort() is called and a debugger can be used to 
 *      investigate the call stack and find the area of code where the pointer
 *      was discarded.
 *
 *    Up(const T *ci_tp)
 *      This method increases the reference count by one, or begins reference
 *      counting if this is the first use of a pointer.
 *
 *    ~RefMgr()
 *      The destructor prints statistics about how many objects of this type
 *      were created and destroyed.  If any remain unallocated, they will
 *      be enumerated.  Ordinarily there are none, as unreferenced objects
 *      should trigger an abort() when the last reference is lost, but 
 *      circular references and other obscure circumstances can occasionally 
 *      slip past.
 *
 *  Properties
 *
 *    RefMap m_fc
 *      A RefMap is typedef'd to mean a map<const T*,int>.  This map holds
 *      all the allocated pointers and their associated reference counts.
 *      TODO: Where supported, a hash_map would give better performance.
 *
 *    unsigned m_uNew
 *      Keeps track of the number of times objects of this type have been
 *      created.
 *
 *    unsigned m_uDelete
 *      Used to keep track of the number of times objects of this type have
 *      been destroyed.
 */


template <class T> class RefMgr {
typedef map<const T*,int> RefMap;
public:
  RefMgr(void) {
    MutexHolder mh(&m_mx);
    cerr << "Initializing map for " << typeid(T).name() << endl;
    m_uNew = 0;
    m_uDelete = 0;
  }
  void Up(const T *i_tp) {
    MutexHolder mh(&m_mx);
    if (i_tp==NULL) return;
    typename RefMap::iterator fci;
    fci = m_fc.find(i_tp);
    if (fci==m_fc.end()) { m_fc[i_tp] = 1; m_uNew++; }
    else {
      (*fci).second++;
      if ((*fci).second==1) m_uNew++; 
    }
  }
  void Down(const T *i_tp) {
    MutexHolder mh(&m_mx);
    if (i_tp==NULL) return;
    typename RefMap::iterator fci;
    fci = m_fc.find(i_tp);
    if (fci==m_fc.end()) abort();
    if ((*fci).second<=0) abort(); 
    (*fci).second--;
    if ((*fci).second==0) abort(); // No references remain.
  }
  void Delete(const T *i_tp) {
    MutexHolder mh(&m_mx);
    if (i_tp==NULL) abort();
    typename RefMap::iterator fci;
    fci = m_fc.find(i_tp);
    if (fci==m_fc.end()) abort();
    if ((*fci).second<=0) abort();
    (*fci).second--;
    if ((*fci).second>0) cerr << typeid(T).name() << " deleted at " << i_tp << " with " << (*fci).second << " references remaining." << endl;
    m_uDelete++;
  }
  ~RefMgr(void) {
    MutexHolder mh(&m_mx);
    typename RefMap::iterator fci;
    cerr << typeid(T).name() << " created " << m_uNew << " deleted " << m_uDelete << endl;
    for (fci=m_fc.begin(); fci!=m_fc.end(); ++fci) {
      if ((*fci).second!=0) {
        cerr << typeid(T).name() << " at " << ((*fci).first)
             << " was not freed." << endl;
      }
    }
  }
  
protected:
  Mutex  m_mx;
  RefMap m_fc;
  unsigned m_uNew;
  unsigned m_uDelete;
};


/*******************
*                  *
*  CLASS DEBUGPTR  *
*                  *
*******************/

/*
 *  The DebugPtr<T> class is a limited form of smart pointer used to track 
 *  pointers to objects of type T.  It uses a matching RefMgr<T> singleton
 *  to store the statistics.
 *
 *  Methods
 *
 *    DebugPtr(void)
 *    DebugPtr(T *i_tp)
 *    DebugPtr(const DebugPtr<T> &ci_tdr)
 *      A DebugPtr<T> can be constructed in one of three ways.  By default, 
 *      it is initialized to NULL.  It can also be initialized with a pointer
 *      to a T object, such as that returned by "new T".  Lastly, it can be
 *      copied from an existing DebugPtr<T> object.  The second and third forms
 *      increment the reference count on the pointer.  In the second case, this
 *      generally means setting it to 1.
 *
 *    operator->(void) const
 *    operator->(void)
 *      The DebugPtr<T> class overloads its -> operator to emulate behavior 
 *      of the contained pointer.  The const overload returns a const pointer.
 *
 *    T &operator*(void) const
 *      The DebugPtr<T> class also overrides its * operator to emulate the
 *      behavior of the contained pointer.  Note that although this method
 *      is const in terms of the DebugPtr and its contents, the reference
 *      returned is not const and may be used to modify the underlying data.
 *    
 *    operator=(T *i_tp)
 *    operator=(const DebugPtr<T> &ci_tdpr)
 *      The = operator is overloaded to allow assignment.  A DebugPtr<T> can
 *      receive an assignment from either a pointer to a T object or from 
 *      another DebugPtr<T>.  In either case, the reference count on the current
 *      value is decremented while the count on the new value is incremented.
 *
 *    operator==(const T *ci_vp) const
 *    operator==(const DebugPtr<T> &ci_tdpr) const
 *    operator==(const DebugPtr<const T> &ci_tdpr) const
 *      The comparison operator (==) is overloaded to compare the underlying
 *      pointer rather than the DebugPtr<T> object itself.  
 *
 *    operator!=(const T *i_tp) const
 *    operator==(const DebugPtr<T> &ci_tdpr) const
 *    operator==(const DebugPtr<const T> &ci_tdpr) const
 *      Returns true if the == equivalent is false.
 *
 *    Delete(void)
 *      This method is called by SoftDelete() to free the DebugPtr<T> and
 *      its contents.  Since C++ operator semantics do not allow overloading
 *      the direct deletion of a non-pointer type, objects that are or are
 *      potentially pointed to by DebugPtr<T> are usually freed with 
 *      SoftDelete().  This alerts the RefMgr<T> that the deletion is occuring
 *      and deletes the underlying pointer.
 *
 *    Get(void)
 *      This method returns the member pointer directly for those situations
 *      that accept no imitations.
 *
 *    Set(const T *i_tp)
 *      This method is used internally by the constructors and the assignment
 *      operator for the common code.  Called externally, it is equivalent to
 *      assignment.
 *
 *    ~DebugPtr()
 *      The destructor decrements the reference count on the contained pointer.
 *      See RefMgr::Down() for a discussion of what happens when the reference
 *      count reaches zero without an explicit delete.
 *      
 *  Properties
 *
 *    T *m_tp
 *      The pointer around which the entire object revolves.
 *
 *  Note: the unquoted word "xfactor" appears in certain methods below.  This
 *  indicates that although the method was coded, it has never been instantiated
 *  by the compiler and thus has never been proven to work correctly.  It should
 *  therefore be treated as something of an unknown. 
 */

template <class T> class DebugPtr {
typedef RefMgr<T> RefMgrT;
public:
  DebugPtr(void) { m_tp = NULL; }
  DebugPtr(T *i_tp) { m_tp = NULL; this->Set(i_tp); }
  DebugPtr(const DebugPtr<T> &ci_tdr) { m_tp=NULL; this->Set(ci_tdr.m_tp); }
  inline const T *operator->(void) const { return m_tp; }
  inline T       *operator->(void) { return m_tp; }
  inline T       &operator*(void) const { return *m_tp; }
  inline void     operator=(T *i_tp) { this->Set(i_tp); }
  inline void     operator=(const DebugPtr<T> &ci_tdpr) { 
    this->Set(ci_tdpr.m_tp); 
  }
  inline bool     operator==(const T *i_tp) const { return (m_tp==i_tp); }
  inline bool     operator==(const DebugPtr<T> &ci_tdpr) const { 
    return (m_tp==i_pv); 
    xfactor;  
  }
  inline bool     operator==(const DebugPtr<const T> &ci_tdpr) const { 
    return (m_tp==ci_tdpr.m_tp); 
    xfactor;  
  }
  inline bool     operator!=(const T *ci_tp) const { return (m_tp!=ci_tp); }
  inline bool     operator!=(const DebugPtr<T> &ci_tdpr) const { 
    return (m_tp!=ci_tdpr.m_tp); 
    xfactor; 
  }
  inline bool     operator!=(const DebugPtr<const T> &ci_tdpr) const { 
    return (m_tp!=ci_tdpr.m_tp); 
    xfactor; 
  }
  inline void     Delete(void) {
    GetSingleton<RefMgrT>()->Delete(m_tp);
    delete m_tp;
    m_tp = NULL;
  }   
  inline const T* Get(void) const { return m_tp; }
  inline void Set(T *i_tp) {
    RefMgrT *rgp = GetSingleton<RefMgrT>();
    rgp->Down(m_tp);
    m_tp = i_tp;
    rgp->Up(m_tp);
  }
  ~DebugPtr(void) {
    RefMgrT *rgp = GetSingleton<RefMgrT>();
    rgp->Down(m_tp);
    m_tp = NULL;
  }
protected:
  mutable T *m_tp;
};


/*
 *  DebugPtr<const T> is a partial template specialization of DebugPtr<T>.
 *
 *  DebugPtr<const T> has all the same methods as DebugPtr<T> with the addition
 *  of methods to support intake of pointers from both DebugPtr<T> and 
 *  DebugPtr<const T> types for construction and assignment.  This models
 *  the behavior of real pointers, where a pointer to an object can
 *  be assigned to a pointer to a const object, but not vice-versa.
 *
 *  This is not necessarily the easiest way to model the desired behavior.  
 *  However, someone recently asked me what partial template specialization
 *  might be useful for "in the real world."
 */


template <class T> class DebugPtr<const T> {
typedef RefMgr<T> RefMgrT;
public:
  DebugPtr(void) { cm_tp = NULL; }
  DebugPtr(const DebugPtr<const T> &ci_tdr) { 
    cm_tp = NULL; 
    this->Set(ci_tdr.Get()); 
  }
  DebugPtr(const DebugPtr<T> i_tdp) { 
    cm_tp = NULL; 
    this->Set(i_tdp.Get());
  }
  DebugPtr(const T *ci_tp) { 
    cm_tp = NULL; 
    this->Set(ci_tp); 
  }
  inline void Set(const T *ci_tp) { 
    RefMgrT *rgp = GetSingleton<RefMgrT>();
    rgp->Down(cm_tp);
    cm_tp = ci_tp;
    rgp->Up(cm_tp);
  }
  inline const T *operator->(void) const { return cm_tp; }
  inline const T &operator*(void) const { return *cm_tp; }
  inline void     operator=(const T *i_tp) { this->Set(i_tp); }
  inline void     operator=(const DebugPtr<const T> &ci_tdp) { 
    this->Set(ci_tdp.cm_tp);
  }
  inline bool     operator==(const void* ci_pv) const { return (cm_tp==ci_pv); }
  inline bool     operator==(const DebugPtr<T> &ci_tdpr) const { 
    return (cm_tp==ci_tdpr.m_tp);  
    xfactor;  
  }
  inline bool     operator==(const DebugPtr<const T> &ci_tdpr) const { 
    return (cm_tp==ci_tdpr.cm_tp); 
    xfactor;   
  }
  inline bool     operator!=(const T* ci_tp) const { return (cm_tp!=ci_tp); }
  inline bool     operator!=(const DebugPtr<T> &ci_tdpr) const { 
    return (cm_tp!=ci_tdpr.m_tp);  
    xfactor; 
  }
  inline bool     operator!=(const DebugPtr<const T> &ci_tdpr) const { 
    return (cm_tp!=ci_tdpr.cm_tp); 
  }
  inline const T* Get(void) const { return cm_tp; }
  ~DebugPtr(void) {
    RefMgrT *rgp = GetSingleton<RefMgrT>();
    rgp->Down(cm_tp);
  }
protected:
  const T *cm_tp;
};
    

#endif  // USE_DEBUG_POINTERS
#endif  // WWWI_MIND_DEBUGPTR_H


